package fitbit

import (
	"context"
	"crypto/hmac"
	"crypto/sha1"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"time"

	"github.com/octo/kraftakt/app"
	"golang.org/x/oauth2"
	oauth2fitbit "golang.org/x/oauth2/fitbit"
	"google.golang.org/appengine/log"
)

func oauthConfig() *oauth2.Config {
	return &oauth2.Config{
		ClientID:     app.Config.FitbitClientID,
		ClientSecret: app.Config.FitbitClientSecret,
		Endpoint:     oauth2fitbit.Endpoint,
		RedirectURL:  "https://kraftakt.octo.it/fitbit/grant",
		Scopes: []string{
			"activity",
			"heartrate",
			"profile",
			"sleep",
		},
	}
}

const csrfToken = "@CSRFTOKEN@"

func AuthURL() string {
	return oauthConfig().AuthCodeURL(csrfToken, oauth2.AccessTypeOffline)
}

func ParseToken(ctx context.Context, r *http.Request, u *app.User) error {
	if state := r.FormValue("state"); state != csrfToken {
		return fmt.Errorf("invalid state parameter: %q", state)
	}

	tok, err := oauthConfig().Exchange(ctx, r.FormValue("code"))
	if err != nil {
		return err
	}

	return u.SetToken(ctx, "Fitbit", tok)
}

func CheckSignature(ctx context.Context, payload []byte, rawSig string) bool {
	signatureGot, err := base64.StdEncoding.DecodeString(rawSig)
	if err != nil {
		log.Errorf(ctx, "base64.StdEncoding.DecodeString(%q) = %v", rawSig, err)
		return false
	}

	mac := hmac.New(sha1.New, []byte(oauthConfig().ClientSecret+"&"))
	mac.Write(payload)
	signatureWant := mac.Sum(nil)

	return hmac.Equal(signatureGot, signatureWant)
}

type Activity struct {
	ActivityID         int       `json:"activityId"`
	ActivityParentID   int       `json:"activityParentId"`
	ActivityParentName string    `json:"activityParentName"`
	Calories           int       `json:"calories"`
	Description        string    `json:"description"`
	Distance           float64   `json:"distance"`
	Duration           int       `json:"duration"`
	HasStartTime       bool      `json:"hasStartTime"`
	IsFavorite         bool      `json:"isFavorite"`
	LastModified       time.Time `json:"lastModified"`
	LogID              int       `json:"logId"`
	Name               string    `json:"name"`
	StartTime          string    `json:"startTime"`
	StartDate          string    `json:"startDate"`
	Steps              int       `json:"steps"`
}

type Distance struct {
	Activity string  `json:"activity"`
	Distance float64 `json:"distance"`
}

type HeartRateZone struct {
	Name        string  `json:"name"`
	Min         int     `json:"min"`
	Max         int     `json:"max"`
	Minutes     int     `json:"minutes"`
	CaloriesOut float64 `json:"caloriesOut"`
}

type ActivitySummary struct {
	Activities []Activity `json:"activities"`
	Goals      struct {
		CaloriesOut int     `json:"caloriesOut"`
		Distance    float64 `json:"distance"`
		Floors      int     `json:"floors"`
		Steps       int     `json:"steps"`
	} `json:"goals"`
	Summary struct {
		ActiveScore          int             `json:"activeScore"`
		ActivityCalories     int             `json:"activityCalories"`
		CaloriesBMR          int             `json:"caloriesBMR"`
		CaloriesOut          float64         `json:"caloriesOut"`
		Distances            []Distance      `json:"distances"`
		Elevation            float64         `json:"elevation"`
		Floors               int             `json:"floors"`
		HeartRateZones       []HeartRateZone `json:"heartRateZones"`
		CustomHeartRateZones []HeartRateZone `json:"customHeartRateZones"`
		MarginalCalories     int             `json:"marginalCalories"`
		RestingHeartRate     int             `json:"restingHeartRate"`
		Steps                int             `json:"steps"`
		SedentaryMinutes     int             `json:"sedentaryMinutes"`
		LightlyActiveMinutes int             `json:"lightlyActiveMinutes"`
		FairlyActiveMinutes  int             `json:"fairlyActiveMinutes"`
		VeryActiveMinutes    int             `json:"veryActiveMinutes"`
	} `json:"summary"`
}

type Subscription struct {
	CollectionType string `json:"collectionType"`
	Date           string `json:"date"`
	OwnerID        string `json:"ownerId"`
	OwnerType      string `json:"ownerType"`
	SubscriptionID string `json:"subscriptionId"`
}

type Client struct {
	fitbitUserID string
	appUser      *app.User
	client       *http.Client
}

func NewClient(ctx context.Context, fitbitUserID string, u *app.User) (*Client, error) {
	if fitbitUserID == "" {
		fitbitUserID = "-"
	}

	c, err := u.OAuthClient(ctx, "Fitbit", oauthConfig())
	if err != nil {
		return nil, err
	}

	return &Client{
		fitbitUserID: fitbitUserID,
		appUser:      u,
		client:       c,
	}, nil
}

func (c *Client) ActivitySummary(ctx context.Context, date string) (*ActivitySummary, error) {
	url := fmt.Sprintf("https://api.fitbit.com/1/user/%s/activities/date/%s.json",
		c.fitbitUserID, date)

	res, err := c.client.Get(url)
	if err != nil {
		return nil, err
	}
	defer res.Body.Close()

	data, _ := ioutil.ReadAll(res.Body)
	log.Debugf(ctx, "GET %s -> %s", url, data)

	var summary ActivitySummary
	if err := json.Unmarshal(data, &summary); err != nil {
		return nil, err
	}

	return &summary, nil
}

func (c *Client) Subscribe(ctx context.Context, collection string) error {
	url := fmt.Sprintf("https://api.fitbit.com/1/user/%s/%s/apiSubscriptions/%s.json",
		c.fitbitUserID, collection, c.appUser.ID)
	res, err := c.client.Post(url, "", nil)
	if err != nil {
		return err
	}
	defer res.Body.Close()

	if res.StatusCode >= 400 && res.StatusCode != http.StatusConflict {
		data, _ := ioutil.ReadAll(res.Body)
		log.Errorf(ctx, "creating %q subscription failed: status %d %q", collection, res.StatusCode, data)
		return fmt.Errorf("creating %q subscription failed", collection)
	}

	return nil
}

func (c *Client) Unsubscribe(ctx context.Context, collection string) error {
	url := fmt.Sprintf("https://api.fitbit.com/1/user/%s/%s/apiSubscriptions/%s.json",
		c.fitbitUserID, collection, c.appUser.ID)
	req, err := http.NewRequest(http.MethodDelete, url, nil)
	if err != nil {
		return err
	}

	res, err := c.client.Do(req.WithContext(ctx))
	if err != nil {
		return err
	}
	defer res.Body.Close()

	if res.StatusCode >= 400 && res.StatusCode != http.StatusNotFound {
		data, _ := ioutil.ReadAll(res.Body)
		log.Errorf(ctx, "deleting %q subscription failed: status %d %q", collection, res.StatusCode, data)
		return fmt.Errorf("deleting %q subscription failed", collection)
	}

	return nil
}

func (c *Client) DeleteToken(ctx context.Context) error {
	return c.appUser.DeleteToken(ctx, "Fitbit")
}

type Profile struct {
	Name     string
	Timezone *time.Location
}

func (c *Client) Profile(ctx context.Context) (*Profile, error) {
	res, err := c.client.Get("https://api.fitbit.com/1/user/-/profile.json")
	if err != nil {
		return nil, err
	}
	defer res.Body.Close()

	if res.StatusCode >= 400 {
		data, _ := ioutil.ReadAll(res.Body)
		log.Errorf(ctx, "reading profile failed: %s", data)
		return nil, fmt.Errorf("HTTP %d error", res.StatusCode)
	}

	var data struct {
		User struct {
			FullName            string
			OffsetFromUTCMillis int
			Timezone            string
		}
	}
	if err := json.NewDecoder(res.Body).Decode(&data); err != nil {
		return nil, err
	}

	loc, err := time.LoadLocation(data.User.Timezone)
	if err != nil {
		loc = time.FixedZone("Fitbit preference", data.User.OffsetFromUTCMillis/1000)
	}

	return &Profile{
		Name:     data.User.FullName,
		Timezone: loc,
	}, nil
}
